/********************************************************
 *
 *  Copyright (c) 2012,  Nothing Real VFX
 *
 *          www.nothing-real.com
 *         belal@nothing-real.com
 *
 *          All rights reserved.
 *
 *  see enclosed LICENSE file for more information
 *
 ********************************************************/


/*
 *   Python Script for:
 *   Generating Renderman declaration files to be used for decl_rioptions.rman and gui.rman
 *
	params = ["filmBack_W", "filmBack_H", "xOffset", "yOffset", "distortion", "anamorphicSqueeze", "xCurvature", "yCurvature", "quarticDist", "dd2", "ud2", "vd2", "qd4", "ud4", "vd4", "d6", "d8", "phi", "B", "cx02d2", "cy02d2", "cx22d2", "cy22d2", "cx04d4", "cy04d4", "cx24d4", "cy24d4", "cx44d4", "cy44d4", "rot", "sqx", "sqy", "cx06d6", "cy06d6", "cx26d6", "cy26d6", "cx46d6", "cy46d6", "cx66d6", "cy66d6"]

	for param in params:
		decl_param=\
		'Declare riopt {float Projection:%s} {\n\
		label "%s"\n\
		subtype slider\n\
		range {-1 1}\n\
		requires Projection:name {bs_tdeLens}\n}    '%(param, param)
		print decl_param

	for param in params:
		defautl = 'riopt Projection:%s {\n    default 0\n}'%param
		print defautl
 */

#include <RixProjection.h>
#include <RixIntegrator.h>
#include <RixShadingUtils.h>

#include <string.h>
#include <fstream>

#include <ldpk/ldpk.h>

using namespace std;
using namespace ldpk;


// This plugin reimplements the standard perspective projection.  It shows
// the details of turning screen and aperture samples into rays (complete
// with derivatives) that are exactly identical to those from the built-in
// perspective projection.  It's meant to provide a good base for tweaking.

class bs_tdeLens : public RixProjection
{

public:
    bs_tdeLens();
    virtual ~bs_tdeLens();

    virtual int Init(
        RixContext &ctx,
        char const *pluginPath );
    virtual RixSCParamInfo const *GetParamTable();
    virtual void Finalize(
        RixContext &ctx );

    virtual void RenderBegin(
        RixContext &ctx,
        RixProjectionEnvironment &env,
        RixParameterList const *parms );
    virtual void RenderEnd(
        RixContext &ctx );

    virtual void Synchronize(
        RixContext &ctx,
        RixSCSyncMsg syncMsg,
        RixParameterList const *parms );

    virtual void Project(
        RixProjectionContext &pCtx );

    virtual void distort(RtFloat *xdn, RtFloat *ydn);


private:
    plugin_loader pl;
    RtFloat lensRadius;
    RtFloat xyStep;
    RtFloat fieldOfView;
    RtFloat zoom;
    RtFloat spread;
    RtFloat maxX;
    RtFloat maxY;

    RtInt doDistort;
    RtInt model;
    RtConstString distFile;
    RtFloat focus_distance;
    RtFloat focal_length;
    RtFloat filmBack_W;
    RtFloat filmBack_H;
    RtFloat xOffset;
    RtFloat yOffset;
    RtFloat pixel_aspect;
    RtFloat distortion;
    RtFloat anamorphicSqueeze;
    RtFloat xCurvature;
    RtFloat yCurvature;
    RtFloat quarticDist;
    RtFloat dd2; RtFloat ud2; RtFloat vd2; RtFloat qd4; RtFloat ud4; RtFloat vd4; RtFloat d6; RtFloat d8;
    RtFloat phi; RtFloat B;
    RtFloat cx02d2; RtFloat cy02d2; RtFloat cx22d2; RtFloat cy22d2; RtFloat cx04d4; RtFloat cy04d4;
    RtFloat cx24d4; RtFloat cy24d4; RtFloat cx44d4; RtFloat cy44d4;
    RtFloat rot; RtFloat sqx; RtFloat sqy;
    RtFloat cx06d6; RtFloat cy06d6; RtFloat cx26d6; RtFloat cy26d6; RtFloat cx46d6; RtFloat cy46d6;
    RtFloat cx66d6; RtFloat cy66d6;
    RtFloat resAspect; RtFloat deviceAspect; RtFloat ratio;
    RixMessages *msg;
};


bs_tdeLens::bs_tdeLens()
{
	// Params hidden from user
	// (computed internally and/or queried from generated Rib based on Maya Camera settings)
    model = 0;
    fieldOfView = 90;
    lensRadius = 5;
    xyStep = 1;
    zoom = 1;
    spread = 1;
    pixel_aspect=1;
    maxX=-1;
    maxY=1;

    // Lens User parameters (need to create for them a UI)
    doDistort = 1;
    distFile="";
    focus_distance = 25;
    focal_length=35;
    filmBack_W=1.417;
    filmBack_H=0.945;
    xOffset=0;
    yOffset=0;
    distortion=0;
    anamorphicSqueeze=1;
    xCurvature=0;
    yCurvature=0;
    quarticDist=0;
    dd2=0; ud2=0; vd2=0; qd4=0; ud4=0; vd4=0; d6=0; d8=0;
    phi=0; B=0;
    cx02d2=0; cy02d2=0; cx22d2=0; cy22d2=0; cx04d4=0; cy04d4=0;
    cx24d4=0; cy24d4=0; cx44d4=0; cy44d4=0;
    rot=0; sqx=1; sqy=1;
    cx06d6=0; cy06d6=0; cx26d6=0; cy26d6=0; cx46d6=0; cy46d6=0;
    cx66d6=0; cy66d6=0;
    resAspect=1; deviceAspect=1; ratio=1;
    msg=NULL;
}

bs_tdeLens::~bs_tdeLens()
{
}

void bs_tdeLens::distort(RtFloat *xdn, RtFloat *ydn)
{
    double x_dn0, y_dn0;
    double x_dn = *xdn;
    double y_dn = *ydn;

    pl.get_model()->undistort(x_dn, y_dn, x_dn0, y_dn0);

    *xdn = x_dn0;
    *ydn = y_dn0;
}


int bs_tdeLens::Init(
    RixContext &ctx,
    char const *pluginPath )
{
	msg = (RixMessages*)ctx.GetRixInterface(k_RixMessages);
	msg->InfoAlways("{INFO}    Initialized bs_tdeLens v1.0");
    return 0;
}

RixSCParamInfo const *bs_tdeLens::GetParamTable()
{
    static RixSCParamInfo s_ptable[] =
    {
        RixSCParamInfo( "fov", k_RixSCFloat ),

        RixSCParamInfo() // end of table
    };
    return &s_ptable[ 0 ];
}

void bs_tdeLens::Finalize(
    RixContext &ctx )
{
}


void bs_tdeLens::RenderBegin(
    RixContext &ctx,
    RixProjectionEnvironment &env,
    RixParameterList const *parms )
{
    RixRenderState &state = *reinterpret_cast< RixRenderState * >(
        ctx.GetRixInterface( k_RixRenderState ) );
    RixRenderState::Type fmtType;
    RtInt resCnt;
    RtInt paramId;

    if ( parms->GetParamId( "filmBack_W", &paramId ) == 0 )
    {
    	parms->EvalParam( paramId, 0, &filmBack_W );
    	filmBack_W = filmBack_W * 2.54f;   // film back cm
    }

    if ( parms->GetParamId( "filmBack_H", &paramId ) == 0 )
    {
    	parms->EvalParam( paramId, 0, &filmBack_H );
    	filmBack_H = filmBack_H * 2.54f;   // film back cm
    }

    struct { RtFloat fstop, flength, fdistance; } dof;
    struct { RtFloat left, right, top, bottom; } window;
    struct { RtFloat width, height, pixelAspect; } format;
    state.GetOption( "Ri:DepthOfField", &dof, 3 * sizeof( RtFloat ),
                     &fmtType, &resCnt );
    state.GetOption( "Ri:ScreenWindow", &window, 4 * sizeof( RtFloat ),
                     &fmtType, &resCnt );
    state.GetOption( "DeviceResolution", &format, 3 * sizeof( RtFloat ),
                     &fmtType, &resCnt );

    // get focal_length and focus_distance from DOF if dof is enabled
    focal_length = dof.flength; 			// Already in cm
    focus_distance = dof.fdistance;			// Already in cm

    pixel_aspect = format.pixelAspect;

    maxX = window.right;
    maxY = window.top;

    lensRadius = 0.5f * dof.flength / dof.fstop;

    float screenWidth = window.right - window.left;
    float screenHeight = window.top - window.bottom;
    if ( screenWidth == 0.0f || screenHeight == 0.0f )
    {
        float aspect = format.width * format.pixelAspect / format.height;
        screenWidth = std::max( 2.0f, 2.0f * aspect );
        screenHeight = std::max( 2.0f, 2.0f / aspect );
    }
    float xStep = (0.25f * screenWidth / format.width);
    float yStep = (0.25f * screenHeight / format.height);
    xyStep = std::max( xStep, yStep );

    Synchronize( ctx, k_RixSCInstanceEdit, parms );
}

void bs_tdeLens::RenderEnd(
    RixContext &ctx )
{
}


void bs_tdeLens::Synchronize(
    RixContext &ctx,
    RixSCSyncMsg syncMsg,
    RixParameterList const *parms )
{
    if ( syncMsg != k_RixSCInstanceEdit )
        return;

    // Evaluate given parameters given in Rib (others remain on initialized defaults):
    RtInt paramId;

    if ( parms->GetParamId( "doDistort", &paramId ) == 0 )
		parms->EvalParam( paramId, 0, &doDistort );

    if ( parms->GetParamId( "model", &paramId ) == 0 )
		parms->EvalParam( paramId, 0, &model );

    if ( parms->GetParamId( "distFile", &paramId ) == 0 )
		parms->EvalParam( paramId, 0, &distFile );

    if ( parms->GetParamId( "xOffset", &paramId ) == 0 )
    {
		parms->EvalParam( paramId, 0, &xOffset );
		xOffset = xOffset * 2.54f;
    }

    if ( parms->GetParamId( "yOffset", &paramId ) == 0 )
    {
		parms->EvalParam( paramId, 0, &yOffset );
		yOffset = yOffset * 2.54f;
    }

    if ( parms->GetParamId( "distortion", &paramId ) == 0 )
		parms->EvalParam( paramId, 0, &distortion );

    if ( parms->GetParamId( "anamorphicSqueeze", &paramId ) == 0 )
		parms->EvalParam( paramId, 0, &anamorphicSqueeze );

    if ( parms->GetParamId( "xCurvature", &paramId ) == 0 )
		parms->EvalParam( paramId, 0, &xCurvature );

    if ( parms->GetParamId( "yCurvature", &paramId ) == 0 )
		parms->EvalParam( paramId, 0, &yCurvature );

    if ( parms->GetParamId( "quarticDist", &paramId ) == 0 )
		parms->EvalParam( paramId, 0, &quarticDist );

    if ( parms->GetParamId( "dd2", &paramId ) == 0 )
		parms->EvalParam( paramId, 0, &dd2 );

    if ( parms->GetParamId( "ud2", &paramId ) == 0 )
		parms->EvalParam( paramId, 0, &ud2 );

    if ( parms->GetParamId( "vd2", &paramId ) == 0 )
		parms->EvalParam( paramId, 0, &vd2 );

    if ( parms->GetParamId( "qd4", &paramId ) == 0 )
		parms->EvalParam( paramId, 0, &qd4 );

    if ( parms->GetParamId( "ud4", &paramId ) == 0 )
		parms->EvalParam( paramId, 0, &ud4 );

    if ( parms->GetParamId( "vd4", &paramId ) == 0 )
		parms->EvalParam( paramId, 0, &vd4 );

    if ( parms->GetParamId( "d6", &paramId ) == 0 )
		parms->EvalParam( paramId, 0, &d6 );

    if ( parms->GetParamId( "d8", &paramId ) == 0 )
		parms->EvalParam( paramId, 0, &d8 );

    if ( parms->GetParamId( "phi", &paramId ) == 0 )
		parms->EvalParam( paramId, 0, &phi );

    if ( parms->GetParamId( "B", &paramId ) == 0 )
    	parms->EvalParam( paramId, 0, &B );

    if ( parms->GetParamId( "cx02d2", &paramId ) == 0 )
		parms->EvalParam( paramId, 0, &cx02d2 );

    if ( parms->GetParamId( "cy02d2", &paramId ) == 0 )
		parms->EvalParam( paramId, 0, &cy02d2 );

    if ( parms->GetParamId( "cx22d2", &paramId ) == 0 )
		parms->EvalParam( paramId, 0, &cx22d2 );

    if ( parms->GetParamId( "cy22d2", &paramId ) == 0 )
		parms->EvalParam( paramId, 0, &cy22d2 );

    if ( parms->GetParamId( "cx04d4", &paramId ) == 0 )
		parms->EvalParam( paramId, 0, &cx04d4 );

    if ( parms->GetParamId( "cy04d4", &paramId ) == 0 )
		parms->EvalParam( paramId, 0, &cy04d4 );

    if ( parms->GetParamId( "cx24d4", &paramId ) == 0 )
		parms->EvalParam( paramId, 0, &cx24d4 );

    if ( parms->GetParamId( "cy24d4", &paramId ) == 0 )
		parms->EvalParam( paramId, 0, &cy24d4 );

    if ( parms->GetParamId( "cx44d4", &paramId ) == 0 )
		parms->EvalParam( paramId, 0, &cx44d4 );

    if ( parms->GetParamId( "cy44d4", &paramId ) == 0 )
		parms->EvalParam( paramId, 0, &cy44d4 );

    if ( parms->GetParamId( "rot", &paramId ) == 0 )
		parms->EvalParam( paramId, 0, &rot );

    if ( parms->GetParamId( "sqx", &paramId ) == 0 )
		parms->EvalParam( paramId, 0, &sqx );

    if ( parms->GetParamId( "sqy", &paramId ) == 0 )
		parms->EvalParam( paramId, 0, &sqy );

    if ( parms->GetParamId( "cx06d6", &paramId ) == 0 )
		parms->EvalParam( paramId, 0, &cx06d6 );

    if ( parms->GetParamId( "cy06d6", &paramId ) == 0 )
    	parms->EvalParam( paramId, 0, &cy06d6 );

    if ( parms->GetParamId( "cx26d6", &paramId ) == 0 )
		parms->EvalParam( paramId, 0, &cx26d6 );

    if ( parms->GetParamId( "cy26d6", &paramId ) == 0 )
		parms->EvalParam( paramId, 0, &cy26d6 );

    if ( parms->GetParamId( "cx46d6", &paramId ) == 0 )
		parms->EvalParam( paramId, 0, &cx46d6 );

    if ( parms->GetParamId( "cy46d6", &paramId ) == 0 )
		parms->EvalParam( paramId, 0, &cy46d6 );

    if ( parms->GetParamId( "cx66d6", &paramId ) == 0 )
		parms->EvalParam( paramId, 0, &cx66d6 );

    if ( parms->GetParamId( "cy66d6", &paramId ) == 0 )
    	parms->EvalParam( paramId, 0, &cy66d6 );

    // in case DOF is disabled, both focal_length and focus_distance will be 0
    // and user in this case should
    // so, assume they are given as params by the user in CM
    if (focal_length == 0)
    {
    	if ( parms->GetParamId( "focal_length", &paramId ) == 0 )
    	{
    		parms->EvalParam( paramId, 0, &focal_length );
    		focal_length = focal_length / 10.0f;	// convert to mm

    		if ( parms->GetParamId( "focus_distance", &paramId ) == 0 )
    			parms->EvalParam( paramId, 0, &focus_distance );		//should be given in cm
    	}
    }

    fieldOfView = 2.0f * atan(filmBack_W / (2.0f * focal_length));

    zoom = tanf( 0.5f* fieldOfView );
	fieldOfView = fieldOfView * F_RADTODEG;
	spread = atanf( zoom * xyStep );

    // ------------------ END PARAM EVAL ----------------------- //


    if (doDistort)
    {
		string modelPath;
		string envLens;

		const char *testEnv = getenv("LENSMODEL_PATH");
		if (testEnv == NULL)
			msg->Error("LENSMODEL_PATH was not set!! Set LENSMODEL_PATH to ldpk plugin models path.");

		envLens = testEnv;

        if (model == 0)
        {
            modelPath = envLens + "/tde4_ldp_classic_3de_mixed" + ".so";
            ifstream ifile(modelPath.c_str());
            if (ifile.fail())
            {
                msg->Error("tde4_ldp_classic_3de_mixed.so is not exist! Make sure you set env LENSMODEL_PATH");
            }
        }
        else if (model == 1)
        {
            modelPath = envLens + "/tde4_ldp_radial_decentered_deg_4_cylindric" + ".so";
            ifstream ifile(modelPath.c_str());
            if (ifile.fail())
            	msg->Error("tde4_ldp_radial_decentered_deg_4_cylindric.so is not exist! Make sure you set env LENSMODEL_PATH");
        }
        else if (model == 2)
        {
            modelPath = envLens + "/tde4_ldp_anamorphic_deg_4_rotate_squeeze_xy" + ".so";
            ifstream ifile(modelPath.c_str());
            if (ifile.fail())
            	msg->Error("tde4_ldp_anamorphic_deg_4_rotate_squeeze_xy.so is not exist! Make sure you set env LENSMODEL_PATH");
        }
        else if (model == 3)
        {
            modelPath = envLens + "/tde4_ldp_radial_deg_8" + ".so";
            ifstream ifile(modelPath.c_str());
            if (ifile.fail())
            	msg->Error("tde4_ldp_radial_deg_8.so is not exist! Make sure you set env LENSMODEL_PATH");
        }
        else if (model == 4)
        {
            modelPath = envLens + "/tde4_ldp_anamorphic_deg_6" + ".so";
            ifstream ifile(modelPath.c_str());
            if (ifile.fail())
            	msg->Error("tde4_ldp_anamorphic_deg_6.so is not exist! Make sure you set env LENSMODEL_PATH");
        }
        else // Disable distortion if an invalid model number was given
        	msg->Error("The distortion model given is invalid!!");


		model_parser mp(false);
		pl.open_plugin(modelPath);

		// Set the parameters that are common for any lens
		pl.get_model()->setParameterValue("tde4_focal_length_cm", focal_length);
		pl.get_model()->setParameterValue("tde4_filmback_width_cm", filmBack_W);
		pl.get_model()->setParameterValue("tde4_filmback_height_cm", filmBack_H);
		pl.get_model()->setParameterValue("tde4_pixel_aspect", pixel_aspect);
		pl.get_model()->setParameterValue("tde4_lens_center_offset_x_cm", xOffset);
		pl.get_model()->setParameterValue("tde4_lens_center_offset_y_cm", yOffset);
		pl.get_model()->setParameterValue("tde4_custom_focus_distance_cm", focus_distance);

		switch (model)
		{
		case 0:
			// Set the paramters that are specific to the 'Classic 3DE distortion' model
			pl.get_model()->setParameterValue("Distortion" , distortion);
			pl.get_model()->setParameterValue("Anamorphic Squeeze", anamorphicSqueeze);
			pl.get_model()->setParameterValue("Curvature X", xCurvature);
			pl.get_model()->setParameterValue("Curvature Y", yCurvature);
			pl.get_model()->setParameterValue("Quartic Distortion", quarticDist);
			break;
		case 1:
			// Set the paramters that are specific to the 'Radial - Standard, Degree 4' model
			pl.get_model()->setParameterValue("Distortion - Degree 2" , dd2);
			pl.get_model()->setParameterValue("U - Degree 2" , ud2);
			pl.get_model()->setParameterValue("V - Degree 2" , vd2);
			pl.get_model()->setParameterValue("Quartic Distortion - Degree 4" , qd4);
			pl.get_model()->setParameterValue("U - Degree 4" , ud4);
			pl.get_model()->setParameterValue("V - Degree 4" , vd4);
			pl.get_model()->setParameterValue("Phi - Cylindric Direction" , phi);
			pl.get_model()->setParameterValue("B - Cylindric Bending" , B);
			break;
		case 2:
			// Set the paramters that are specific to the 'Anamorphic - Standard, Degree 4' model
			pl.get_model()->setParameterValue("Cx02 - Degree 2", cx02d2);
			pl.get_model()->setParameterValue("Cy02 - Degree 2", cy02d2);
			pl.get_model()->setParameterValue("Cx22 - Degree 2", cx22d2);
			pl.get_model()->setParameterValue("Cy22 - Degree 2", cy22d2);
			pl.get_model()->setParameterValue("Cx04 - Degree 4", cx04d4);
			pl.get_model()->setParameterValue("Cy04 - Degree 4", cy04d4);
			pl.get_model()->setParameterValue("Cx24 - Degree 4", cx24d4);
			pl.get_model()->setParameterValue("Cy24 - Degree 4", cy24d4);
			pl.get_model()->setParameterValue("Cx44 - Degree 4", cx44d4);
			pl.get_model()->setParameterValue("Cy44 - Degree 4", cy44d4);
			pl.get_model()->setParameterValue("Lens Rotation", rot);
			pl.get_model()->setParameterValue("Squeeze-X", sqx);
			pl.get_model()->setParameterValue("Squeeze-Y", sqy);
			break;
		case 3:
			// Set the paramters that are specific to the 'Radial - Fisheye, Degree 8' model
			pl.get_model()->setParameterValue("Distortion - Degree 2", dd2);
			pl.get_model()->setParameterValue("Quartic Distortion - Degree 4", qd4);
			pl.get_model()->setParameterValue("Degree 6" , d6);
			pl.get_model()->setParameterValue("Degree 8" , d8);
			break;
		case 4:
			pl.get_model()->setParameterValue("Cx02 - Degree 2", cx02d2);
			pl.get_model()->setParameterValue("Cy02 - Degree 2", cy02d2);
			pl.get_model()->setParameterValue("Cx22 - Degree 2", cx22d2);
			pl.get_model()->setParameterValue("Cy22 - Degree 2", cy22d2);
			pl.get_model()->setParameterValue("Cx04 - Degree 4", cx04d4);
			pl.get_model()->setParameterValue("Cy04 - Degree 4", cy04d4);
			pl.get_model()->setParameterValue("Cx24 - Degree 4", cx24d4);
			pl.get_model()->setParameterValue("Cy24 - Degree 4", cy24d4);
			pl.get_model()->setParameterValue("Cx44 - Degree 4", cx44d4);
			pl.get_model()->setParameterValue("Cy44 - Degree 4", cy44d4);
			pl.get_model()->setParameterValue("Cx06 - Degree 6", cx06d6);
			pl.get_model()->setParameterValue("Cy06 - Degree 6", cy06d6);
			pl.get_model()->setParameterValue("Cx26 - Degree 6", cx26d6);
			pl.get_model()->setParameterValue("Cy26 - Degree 6", cy26d6);
			pl.get_model()->setParameterValue("Cx46 - Degree 6", cx46d6);
			pl.get_model()->setParameterValue("Cy46 - Degree 6", cy46d6);
			pl.get_model()->setParameterValue("Cx66 - Degree 6", cx66d6);
			pl.get_model()->setParameterValue("Cy66 - Degree 6", cy66d6);
			break;
		}
		pl.get_model()->initializeParameters();
		// Parsing Distortion Parameters from file if one exists,
		// This overrides any local parameters.
		// I found this useful in many cases and makes consistent workflow between Arnold and 3DE
		// Also the one could use it as a lens Specs file for other non-tracked scenes.
		//const char * distFile = distFile;
		string df = distFile;
		if (df.length() > 1) // null makes some troubles inside Maya if a file was entered and deleted
					// here, a distortion filename length/path should be greater than one character.
		{
			ifstream dFile(distFile);
			if(dFile.fail())
			{
				msg->Error("Cannot open the Distortion parameters file! Local parameters will be used instead.");
			}
			else
			{
				// We get the Common Lens parameters from local parameters if were not been set
				// in the param file.
				msg->Info("We have got a param file.");
				mp.set_model(pl.get_model());
				mp.parse(dFile);
			}
		}
    }
}


void bs_tdeLens::Project(
    RixProjectionContext &pCtx )
{
    for ( int index = 0; index < pCtx.numRays; ++index )
    {
        RtPoint2 &screen( pCtx.screen[ index ] );
        RtPoint2 &aperture( pCtx.aperture[ index ] );
        RtRayGeometry &ray( pCtx.rays[ index ] );

        if (doDistort)
        {
			// normalize screen window from (-min to +max) to (0 to 1)
			screen.x = (screen.x + maxX) / (2.0f * maxX);
			screen.y = (screen.y + maxY) / (2.0f * maxY);

			distort(&screen.x, &screen.y);

			// inverse the normalization function to reformat pdn form -min to +max
			screen.x = (screen.x - 0.5f) * 2.0f * maxX;
			screen.y = (screen.y - 0.5f) * 2.0f * maxY;

			screen.x += (xOffset);
			screen.y += (yOffset);
        }

        ray.origin = RtPoint3(
            aperture.x * lensRadius,
            aperture.y * lensRadius,
            0.0f );

        ray.direction = RtVector3(
            screen.x * zoom * focus_distance - ray.origin.x,
            screen.y * zoom * focus_distance - ray.origin.y,
            focus_distance );

        ray.direction.Normalize();
        ray.originRadius = 0.0f;
        ray.raySpread = spread *ray.direction.z;
    }
}


RIX_PROJECTIONCREATE
{
    return new bs_tdeLens();
}

RIX_PROJECTIONDESTROY
{
    delete reinterpret_cast< bs_tdeLens * >( projection );
}
